#pragma once
/** @file DeviceThreadHelper.h
	Class definitions for DeviceThreadHelper, its subclasses, and utility structures
*/

#include <deque>
#include <vector>
#include "common.h"

class BuildContext;
class Device;

/// A Work packet. What this is depends on the device in question, but it should contain data pointing to a quanta of work
struct DeviceThreadWork {
	int order;
};

/// Thread Data. Each active thread will instantiate a class derived from this, and pass to the Action and Merge funcs
struct DeviceThreadData {
};

/** Framework class to allow your device to more easily perform multithreaded processing

This class fits neatly into @ref page_threading "World Machine's threading model", implemented on a per-device basis
to multithread the processing of all slow devices that are amenable to divided workloads.

If you have an algorithm that works scanline-by-scanline, you should derive from ThreadHelperScanline which provides
incredibly simple multithreading power from only a single overriden function.

How to Use:
	-# Decide how to split up the work in your algorithm (see note above about a better scanline based approach)
		- There is a built-in work queue that you can leverage if your code is amenable to it
		- If you can't use the work queue, you will have to provide your own way of splitting and recombining tasks between threads
	-# Derive a new subclass from DeviceThreadHelper
		- provide a way for your derived class to access your needed datapackets, such as passing them into the constructor
		- Set noAutoReporting in your constructor if you want the framework to NOT automate progress reporting
		- Implement doThreadDataCreate() and doThreadDataDestroy() if you need per-thread data
		- Implement doActionFunc() to process a single packet of work
		- Implement doMergeFunc() to handle integrating a finished packet of work into your end-result
		- If using the work queue:
			- Implement doCreateWork() to create all of your own work packets that will be handed to doActionFunc()
			- Implement doDeleteWork() which will be called after a work packet is finished
			- set order_returning_work to true either directly or through the parent constructor if you cannot accept out-of-order returning work packets
		- If using your own method:
			- Implement workAvailable() and make it return the appropriate value if there is still work to be done. The built-in work packet system will pass
			  NULL values to your doActionFunc() for the work parameter, so you will have to manage everything yourself.
	-# Derive new subclasses from DeviceThreadData and DeviceThreadWork to represent any per-thread data and work packet data. This is almost always necessary, as at a minimum
	   you will need a way of deciding what actions to perform in doActionFunc() when it receives a work packet.
	-# In your device Activate() function, instantiate an object from your ThreadHelperScanline-derived class
	-# Call PerformWork(). As many threads as supported will be launched. Function returns when all tasks are completed.
	-# The results are either directly produced or copied out as results depending on how you've created your subclass.

Progress Reporting:

 DeviceThreadHelper will also call context.ReportProgress() based upon the progress through the worklist, freeing your work function
 from directly concerning itself with this. The framework also WILL SOON handle parent cancel messages correctly for you!

@ingroup threading devices
@deprecated DeviceThreadHelper is no longer recommended for new projects as it is far more complicated and cumbersome than the new C++11 style visitors.
 */
class DeviceThreadHelper
{
public:
	DeviceThreadHelper(Device *_owner, BuildContext &_context, bool ordered);
	~DeviceThreadHelper(void);

	/// Begins threaded device build. This is the only outside-accessible function!
	bool PerformWork();	//Returns only when built results are finished and merged, or user abort
						// (return value of true indicates successful build)
protected:
	BuildContext &context;
	Device *owner;

	bool noAutoReporting;	///< Set to TUE in your constructor if helper should NOT automatically call context.ReportProgress()

/// @name Work Functions. 
/// If you're using the provided work que, fill the queue using doCreateWork(). 
//@{
	virtual void doCreateWork() { 		progress_max = (int)waiting_work.size(); }; // Implement these two functions if noWorkQue == false
	// called when the given work item is being deleted.
	virtual void doDeleteWork(DeviceThreadWork *work) { };
	virtual bool workAvailable();
//@}

/// @name Action Operations -- Implement these for your device!
//@{
	virtual bool doActionFunc(DeviceThreadWork *work, DeviceThreadData *data) = 0;	// Perform build operation on a piece of work
	virtual bool doMergeFunc(DeviceThreadWork *work, DeviceThreadData *data) = 0;	// Merge results of work into final data
//@}

/// @name Local Data Management Functions
/// Create Local data on a per-thread basis. Override DeviceThreadData to define data and create/destroy it per-thread
//@{
	virtual DeviceThreadData *doThreadDataCreate() { return NULL; };
	virtual void			  doThreadDataDestroy(DeviceThreadData *data) { return; };
//@}

/// Functions to call from derived classes
	DeviceThreadWork *RetrieveWork(DeviceThreadData *data);
	bool StoreWork(DeviceThreadWork *work, DeviceThreadData *data);
	void addWorklist(DeviceThreadWork *work);	///< Call from doCreateWork to add items to the worklist
	bool is_reordering() { return order_returning_work; };
	virtual bool threadProgress(int progress);

	int progress_max;
	volatile bool cancel;			// set to true 
	volatile bool error;
	volatile int progress_cur;
	std::deque<DeviceThreadWork*> waiting_work;
	bool order_returning_work;

private: // Implementation details..
	volatile LONG num_threads;		// # of set threads
	volatile LONG thread_counter;	// # of active threads as measured by them
	static UINT staticThreadFunc(void *classptr);

	virtual int ThreadFunc(bool leader, bool allow_childspawn = false);
	CRITICAL_SECTION critsec;
	std::vector<DeviceThreadWork*> returning_work;
	int returning_waitfor;
	void RetireWaitingWork(DeviceThreadData *data);
};

struct ThreadScanlineWork : public DeviceThreadWork {
	Coord range;
};

/** Scanline-based multithreading helper
 ThreadHelperScanline is a specialization for devices that do scanline-based processing, and can thus split up work on a single heightfield
 among threads and work on the same data without clobbering each other. This is especially nice because we get no-contention multithreading,
 and if you setup your subclass so that you have a seperate result packet that is built-up from acting on the source, it is very
 multithreading cache friendly.
 
 All that is necessary to implement from this base class is the doActionFunc method.

 How to Use:

	-# Derive a new subclass from ThreadHelperScanline
		- provide a way for your derived class to access your needed datapackets, such as passing them into the constructor
		- override doActionFunc() and do your processing on the scanline range given by parameter data->range
	-# In your device Activate() function, instantiate an object from your ThreadHelperScanline-derived class
	-# Call setScanlineRange() to the appropriate values
	-# Call PerformWork()
	-# The results are either directly produced or copied out as results depending on how you've created your subclass.

@ingroup threading
*/
class ThreadHelperScanline : public DeviceThreadHelper {
public:
	ThreadHelperScanline(Device *_owner, BuildContext &_context, int _sections = 48);
	/// Call before performing work to set the scanline range to process
	void setScanlineRange(int min, int max) { total_range.x = min; total_range.y = max; };

protected:
	/// Creates scanline-based work data
	virtual void doCreateWork();
	/// Cleans up scanline work data
	virtual void doDeleteWork(DeviceThreadWork *work) { ThreadScanlineWork *trw = (ThreadScanlineWork *) work; delete trw; };
	/// Merge occurs automatically since we're working on same packet but different scanlines
	virtual bool doMergeFunc(DeviceThreadWork *work, DeviceThreadData *data) { return true; };

	/// * This is the function that your derived class must implement! *
	virtual bool doActionFunc(DeviceThreadWork *work, DeviceThreadData *data) = 0;	// Perform build operation on a piece of work

	Coord total_range;
	int sections;
};

/***
A Further extension to the ThreadHelperScanline, ScanlinePhased allows you to easily perform algorithms that are not simple one-pass but rather
have a complicated series of steps that must be followed, but that can nonetheless be broken down into scanlines within each step.

Implement the doActionScanline and doActionPhase

*/
class ThreadHelperScanlinePhased : public ThreadHelperScanline {
public:
	ThreadHelperScanlinePhased(Device *_owner, BuildContext &_context, Coord &range);

protected:
	virtual bool doActionFunc(DeviceThreadWork *work, DeviceThreadData *data);

	/// You must Implement this function! 
	virtual void doActionScanline(int y, DeviceThreadData *data) = 0;	// Perform work on a single scanline!
//	virtual void doActionInterPhase(int phase, DeviceThreadData *data) = 0; 

	int current_phase;
	int num_phase;
};