#pragma once
namespace WorldMachine {

/// @file WMThreadPool.h Abstract Interface for WM Thread Pool

/// User-specified data for the WMThreadPool class
typedef void* WMThreadDataP;
/// User-specified thread function given to WMThreadPool::runThread()
typedef UINT (*WMThreadProc)(WMThreadDataP userdata);

class WMThreadPool;

/// Internal Structure for Thread Pooling implementation
struct WMThreadInternalStruc {
	WMThreadInternalStruc() { owner = NULL; alive = false; idle = false; userproc = NULL; userdata = NULL; thread_event = NULL; };
	WMThreadPool *owner;	// Pointer to the thread pool

	volatile bool alive;	// is there a thread associated with this data?
	volatile LONG idle;		// set to true by the thread if its waiting for a new client function
	WMThreadProc	 userproc;	
	WMThreadDataP	 userdata;
	HANDLE	thread_event;	// signal'd by the thread pool when a new task is recieved.
};

const int WM_THREADMAX = 32;

/** Thread Pooling implementation handles providing available worker threads to clients as requested.
	@ingroup threading

Part of World Machine's @ref threading "threading model". 

On Program Start, the WM app core will initialize the threadpool. 

Clients (devices or any other WM code) may run a worker thread by simply requesting one from the globally-accessible pool with runThread()...
If no more threads can be dispatched at the moment due to all workers being busy, runThread will return false.

Each thread is suspended until provided a procedure. When that client function returns, it goes back to an idle.

Note that WMThreadPool is the abstract base class, to be derived from by the host application.
*/
class WMThreadPool
{
public:
	bool startThreadPool(int num_threads, int thread_priority);
	bool endThreadPool(bool return_immediately = false);

	int maxThreads() { return max_threads; };
	int activeThreads();
	int availableThreads() { return maxThreads() - activeThreads(); };

	virtual bool beginLock() = 0; // if you have several things to do (like start multiple threads)
	virtual bool endLock() = 0;	  // ...

	/// Request a thread from the thread pool to run with the given procedure and data. 
	/// Returns false if thread could not be launched (for example if all threads are currently allocated)
    virtual bool runThread(WMThreadProc userproc, WMThreadDataP userdata) = 0;

protected:
	virtual void startThread(WMThreadInternalStruc *thread_control, int thread_priority) = 0;

	int max_threads;
	int thread_priority;
	WMThreadInternalStruc pool[WM_THREADMAX];

	HANDLE	eventQuit;
	HANDLE	eventDispatch;

	WMThreadPool(void);
	~WMThreadPool(void);
};
}